#include <Arduino.h>
#include "BluetoothSerial.h"
#include <constants.h>
#include <Preferences.h>

Preferences preferences;
#if !defined(CONFIG_BT_ENABLED) || !defined(CONFIG_BLUEDROID_ENABLED)
#error Bluetooth is not enabled! Please run `make menuconfig` to and enable it
#endif

BluetoothSerial SerialBT;

void btMessage(){ // mensaje que se manda por bluetooth con los datos para la prueba de fatiga
  String message=String(pFatiga);
  message+=",";
  message+=String(timeO);
  message+=",";
  message+=String(pot_a);
  message+=",";
  message+=String(angulo_a);
  message+=",";
  message+=String(timeC);
  message+=",";
  message+=String(pot_c);
  message+=",";
  message+=String(angulo_c);
  SerialBT.println(message);
}

void anguloDedo(struct Dedo * dedo){ // función para el cáculo del angulo en el que se encuentran cada uno de los dedos
  dedo->potValor=analogReadMilliVolts(dedo->potPin); //lectura del pin
  dedo->angulo=(62/(dedo->potMax-dedo->potMin))*(dedo->potValor-dedo->potMin); //ecuación 
  
}

void calibDelay(struct Dedo *dedo){ //funcion para la corrección de los delays
  if(dedo->bandDelay==0)
  {
    dedo->potValor=analogReadMilliVolts(dedo->potPin); //lectura del valor del potenciometro usada para ver si el dedo ya finalizo el movimiento
    anguloDedo(dedo); //calculo del ángulo
    switch(dedo->calibracionDelay) //MEF
    { 
      case abrirTodo:
        if (dedo->prevCalibracionDelay != dedo->calibracionDelay){ //se comprueba que estaba en un estado diferente
          Serial.print("Switching to state: ");
          Serial.println(dedo->calibracionDelay);
          dedo->prevCalibracionDelay = dedo->calibracionDelay; //se actualiza el estado
          t_estDelay=millis(); //se captura el tiempo en millis
          dedo->timerDelay1=millis(); //se guarda el tiempo en el que empiueza el movimiento
          analogWrite(dedo->m2,PWM); //se acciona el motor
        }
        if(dedo->angulo>15 && dedo->angulo<16 ){ //cuando esta entre 15° y 16° se toma el tiempo que se demoro
          dedo->timerDelay2=millis();

        }
        if(dedo->potValor-dedo->potAnterior<=10){//se compara si hubo cambios en el potenciometro para determinar si ya termino el movimiento
          
          delay(1);
          if(millis()-t_estDelay>1500){ //si ya ha pasado mas de un segundo se calcula el delay total
            dedo->delayOpen=dedo->timerDelay2-dedo->timerDelay1;
            preferences.putInt((dedo->delayOpenPreferences).c_str(),dedo->delayOpen);//se guarda en preferences
            dedo->calibracionDelay=delayTime; //siguiente estado dentro de la MEF
          } 
        } else{
          t_estDelay=millis(); //si no se cumple la función se actualiza el tiempo
        }
        dedo->potAnterior=dedo->potValor;//se actuasliza el valor del potenciometro
        break;
    
      case delayTime:
        analogWrite(dedo->m1,0);
        analogWrite(dedo->m2,0);
        if (millis() - dedo->timerDedo > 500){ //pequeño delay
          Serial.print("Delay case");
          dedo->calibracionDelay = cerrarTodo;//siguiente estado dentro de la MEF
          }
        break;
      case cerrarTodo:
        if (dedo->prevCalibracionDelay != dedo->calibracionDelay){
          Serial.print("Switching to state: ");
          Serial.println(dedo->calibracionDelay);
          //Serial.print("Cerrar del todo ");
          dedo->prevCalibracionDelay = dedo->calibracionDelay;
          t_estDelay=millis();
          dedo->timerDelay1=millis();
          analogWrite(dedo->m1,PWM);  
        }
        if(dedo->angulo>50 && dedo->angulo<55 ){
          dedo->timerDelay2=millis();
          
        }
        if(abs(dedo->potAnterior-dedo->potValor)<=10){
          delay(1);
          if(millis()-t_estDelay>1000){
            dedo->delayClose=dedo->timerDelay2-dedo->timerDelay1;
            //Serial.print("close Delay: ");
            //Serial.println(dedo->delayClose);
            preferences.putInt((dedo->delayClosePreferences).c_str(),dedo->delayClose);
           
          }
        } else{t_estDelay=millis();}
      
        dedo->potAnterior=dedo->potValor;
        if(dedo->delayOpen!=0 &&  dedo->delayClose!=0){
          dedo->bandDelay=1;
          analogWrite(dedo->m1,0);
          analogWrite(dedo->m2,0);
          
          Serial.println("calibration delay done");
        }
       
        break;
    }
  }
}

void setPWM(struct Dedo * dedo1,struct Dedo * dedo2,struct Dedo * dedo3,struct Dedo * dedo4){
  anguloDedo(dedo1);//calculo del angulo para todos los dedos
  anguloDedo(dedo2);
  anguloDedo(dedo3);
  anguloDedo(dedo4);

  promedioAng=(dedo1->angulo + dedo2->angulo + dedo3->angulo + dedo4->angulo)/4; //promedio
  Serial.print("promedio lectura angulo: ");
  Serial.println(promedioAng);
  Serial.print("angulo 1: ");
  Serial.println(dedo1->angulo);

  Serial.print("angulo 2: ");
  Serial.println(dedo2->angulo);

  Serial.print("angulo 3: ");
  Serial.println(dedo3->angulo);

  Serial.print("angulo 4: ");
  Serial.println(dedo4->angulo);

  //actualización del PWM para cada dedo
 // DEDO 1
  //float resta=promedioAng-(dedo1->angulo);
  //Serial.print("resta: ");
  //Serial.println(resta);
  if((promedioAng-dedo1->angulo)>tol){ //positivo
    if(S_A==true && S_C==false){// esta abriendo
      dedo1->pwm+=stepPot;
      if(dedo1->pwm>250){dedo1->pwm=245;}
    }
    
    else if(S_A==false && S_C==true){
      dedo1->pwm-=stepPot;
      if(dedo1->pwm<100){dedo1->pwm=100;}
    }
  }

  if(promedioAng-dedo1->angulo<tol){//negativo
    if(S_A==true && S_C==false){// esta abriendo
      dedo1->pwm-=stepPot;
      if(dedo1->pwm<100){dedo1->pwm=100;}
    }
    
    if(S_A==false && S_C==true){
      dedo1->pwm+=stepPot;
      if(dedo1->pwm>250){dedo1->pwm=245;}
    }
  }
  //DEDO 2
  if(promedioAng-dedo2->angulo>tol){ //positivo
    if(S_A==true && S_C==false){// esta abriendo
      dedo2->pwm+=stepPot;
      if(dedo2->pwm>250){dedo2->pwm=245;}
    }
    
    if(S_A==false && S_C==true){
      dedo2->pwm-=stepPot;
      if(dedo2->pwm<100){dedo2->pwm=100;}
    }
  }

  else if(promedioAng-dedo2->angulo<tol){//negativo
    if(S_A==true && S_C==false){// esta cerrando
      dedo2->pwm-=stepPot;
      if(dedo2->pwm<100){dedo2->pwm=100;}
    }
    
    if(S_A==false && S_C==true){
      dedo2->pwm+=stepPot;
      if(dedo2->pwm>250){dedo2->pwm=245;}
    }
  }
  //DEDO 3
  if(promedioAng-dedo3->angulo>tol){ //positivo
    if(S_A==true && S_C==false){// esta abriendo
      dedo3->pwm+=stepPot;
      if(dedo3->pwm>250){dedo3->pwm=245;}
    }
    
    if(S_A==false && S_C==true){
      dedo3->pwm-=stepPot;
      if(dedo3->pwm<100){dedo3->pwm=100;}
    }
  }

  else if(promedioAng-dedo3->angulo<tol){//negativo
    if(S_A==true && S_C==false){// esta cerrando
      dedo3->pwm-=stepPot;
      if(dedo3->pwm<100){dedo3->pwm=100;}
    }
    
    if(S_A==false && S_C==true){
      dedo3->pwm+=stepPot;
      if(dedo3->pwm>250){dedo3->pwm=245;}
    }
  }
 // DEDO 4
  if(promedioAng-dedo4->angulo>tol){ //positivo
    if(S_A==true && S_C==false){// esta abriendo
      dedo4->pwm+=stepPot;
      if(dedo4->pwm>250){dedo4->pwm=245;}
    }
    
    if(S_A==false && S_C==true){
      dedo4->pwm+=stepPot;
      if(dedo4->pwm<100){dedo4->pwm=100;}
    }
  }

  else if(promedioAng-dedo4->angulo<tol){//negativo
    if(S_A==true && S_C==false){// esta cerrando
      dedo4->pwm-=stepPot;
      if(dedo4->pwm<100){dedo4->pwm=100;}
    }
    
    if(S_A==false && S_C==true){
      dedo4->pwm+=stepPot;
      if(dedo4->pwm>250){dedo4->pwm=245;}
    }
  }

  }

void abreDedo(struct Dedo * dedo){ //funcion de abrir dedo
  if (dedo->prevEstado != dedo->estadoDedo){ //se comprueba que llegara de un estado diferente
          Serial.print("Switching to state: ");
          Serial.println(dedo->estadoDedo);
          dedo->prevEstado = dedo->estadoDedo; //se actualiza el estado
          timeOpen=millis();//se guarda el tiempo 
        }
  if(millis()-timeOpen<dedo->delayOpen){ //se compara si no ha pasado un tiempo mayor al delay calculado en la funcion anterior
    analogWrite(dedo->m2,0); // se deja el motor inactivo
    
  }
  else{
    analogWrite(dedo->m2,dedo->pwm); //se acciona el motor 
    anguloDedo(dedo); //se calcula el ángulo del dedo
    if(dedo->angulo>=58){// si ha llegado a un angulo cerca al máximo se apaga el motor
      dedo->estadoDedo=1;// bandera que cambia de estado cuando ya ha terminado el movimiento
      analogWrite(dedo->m2,0);
    }
  }
}

void cierraDedo(struct Dedo * dedo){
  if (dedo->prevEstado != dedo->estadoDedo){
          Serial.print("Switching to state: ");
          Serial.println(dedo->estadoDedo);
          dedo->prevEstado = dedo->estadoDedo;
          timeClose=millis();
        }

  if(millis()-timeClose<dedo->delayClose){
    analogWrite(dedo->m1,0);
    //Serial.println("no supera el delay close");
  }
  else{
    analogWrite(dedo->m1,dedo->pwm);
    anguloDedo(dedo);
    //Serial.println("supera el delay close");
    if(dedo->angulo<=5){
      dedo->estadoDedo=2;
      analogWrite(dedo->m1,0);
      }
  }
}

void dedoQuieto(struct Dedo * dedo){

  analogWrite(dedo->m1,0); //se dejan los motores inactivos
  analogWrite(dedo->m2,0);
}

void Calibracion(struct Dedo * dedo){ // funcion de calibración sigue la misma lógica de la función anterior
  if(dedo->band==0)
  {
    dedo->potValor=analogReadMilliVolts(dedo->potPin);
    switch(dedo->calibracion)
    {
      case abrirTodo:
        if (dedo->prevCalibracion != dedo->calibracion){
          Serial.println("Calibracion");
          Serial.print("Switching to state: ");
          Serial.println(dedo->calibracion);
          dedo->prevCalibracion = dedo->calibracion;
          t_est=millis();
          analogWrite(dedo->m2,PWM);
        }
        if(dedo->potValor-dedo->potAnterior<=10){

          int restaTime=millis()-t_est;
          delay(100);
          if(restaTime>1500){
            dedo->potMax=dedo->potValor;
            Serial.print("PotMax:");
            Serial.println(dedo->potMax); //se guarda el valor de milivoltios máximos
            preferences.putInt((dedo->potMaxPreferences).c_str(),dedo->potMax); //se guarda en preferences
            analogWrite(dedo->m1,0);
            analogWrite(dedo->m2,0);
            dedo->timerDedo=millis();
            dedo->calibracion=delayTime;//cambio de estado
          } 
        } 
        else{
          t_est=millis();
        }
        dedo->potAnterior=dedo->potValor;
        break;
    
      case delayTime:
        if (millis() - dedo->timerDedo > 500){
          dedo->calibracion = cerrarTodo;//canbio de estado
          }
        break;
      case cerrarTodo:
        if (dedo->prevCalibracion != dedo->calibracion){
          Serial.print("Switching to state: ");
          Serial.println(dedo->calibracion);
          dedo->prevCalibracion = dedo->calibracion;
          t_est=millis();
          analogWrite(dedo->m1,PWM);  
        }
        if(abs(dedo->potAnterior-dedo->potValor)<=10){
          delay(100);
          if(millis()-t_est>1000){
            dedo->potMin=dedo->potValor;
            Serial.print("PotMin:");
            Serial.println(dedo->potMin);
            preferences.putInt((dedo->potMinPreferences).c_str(),dedo->potMin);//se guarda en preferences
          }
        } else{t_est=millis();}
      
        dedo->potAnterior=dedo->potValor;
        if(dedo->potMax!=0 && dedo->potMin!=0){
          dedo->band=1;
          analogWrite(dedo->m1,0);
          analogWrite(dedo->m2,0);
          
          Serial.println("calibration done");
        }
        break;
    }
  }
}

void correccionDelay(struct Dedo * dedo1,struct Dedo * dedo2,struct Dedo * dedo3,struct Dedo * dedo4){
  // resta del valor de delay mayor al resto de los delay para el movimiento de abrir
  dedo1->delayOpen=abs(dedo1->delayOpen-dedo3->delayOpen);
  dedo2->delayOpen=abs(dedo2->delayOpen-dedo3->delayOpen);
  dedo4->delayOpen=abs(dedo4->delayOpen-dedo3->delayOpen);
  dedo3->delayOpen=abs(dedo3->delayOpen-dedo3->delayOpen);
  //se guardan los valores en preferences
  preferences.putInt((dedo1->delayOpenPreferences).c_str(),dedo1->delayOpen);
  preferences.putInt((dedo2->delayOpenPreferences).c_str(),dedo2->delayOpen);
  preferences.putInt((dedo3->delayOpenPreferences).c_str(),dedo3->delayOpen);
  preferences.putInt((dedo4->delayOpenPreferences).c_str(),dedo4->delayOpen);

  // resta del valor de delay mayor al resto de los delay para el movimiento de cerrar
  dedo1->delayClose=abs(dedo1->delayClose-dedo4->delayClose);
  dedo2->delayClose=abs(dedo2->delayClose-dedo4->delayClose);
  dedo3->delayClose=abs(dedo3->delayClose-dedo4->delayClose);
  dedo4->delayClose=abs(dedo4->delayClose-dedo4->delayClose);
  //se guardan los valores en preferences
  preferences.putInt((dedo1->delayClosePreferences).c_str(),dedo1->delayClose);
  preferences.putInt((dedo2->delayClosePreferences).c_str(),dedo2->delayClose);
  preferences.putInt((dedo3->delayClosePreferences).c_str(),dedo3->delayClose);
  preferences.putInt((dedo4->delayClosePreferences).c_str(),dedo4->delayClose);

}

void setup() {
  Serial.begin(115200); //Inicio de puerto Serial a 115200 baudios
  SerialBT.begin("prueba fatiga"); //Bluetooth device name
  preferences.begin("config",false);
  pinMode(potMenique, INPUT);
  pinMode(menique1, OUTPUT);
  pinMode(menique2, OUTPUT);
  pinMode(potAnular, INPUT);
  pinMode(anular1, OUTPUT);
  pinMode(anular2, OUTPUT);
  pinMode(potMedio, INPUT);
  pinMode(medio1, OUTPUT);
  pinMode(medio2, OUTPUT);
  pinMode(potIndice, INPUT);
  pinMode(indice1, OUTPUT);
  pinMode(indice2, OUTPUT);
  
  delay(10000); //para alcanzar a abrir Serial monitor
  Serial.println("Init");
  initTime=millis();
  
  bandOk=preferences.getInt("bandCalib",bandOk);//se obtiene la bandera de calibración
  Serial.print("bandOk: ");
  Serial.print(bandOk);
  
  bandDelay=preferences.getInt("bandDelay",bandDelay);//se obtiene la bandera de corrección de delay
  Serial.print("bandDelay: ");
  Serial.print(bandDelay);

  while(bandOk==0){ //se haci la calibración
    Calibracion(&indice);
    Calibracion(&medio);
    Calibracion(&anular);
    Calibracion(&menique);
    if(indice.band==1 && medio.band==1 && anular.band==1 && menique.band==1){ //cuando todos los dedos terminan de calibrar se cumple esta condición
      bandOk=1;//bandera de calibración
      preferences.putInt("bandCalib",bandOk); //se guarda en preferences
      Serial.print("bandOk: ");
      Serial.println(bandOk);
    }
    
  }

  while(bandDelay==0){ //corrección de delay
    calibDelay(&indice);
    calibDelay(&medio);
    calibDelay(&anular);
    calibDelay(&menique);
    if(indice.bandDelay==1 && medio.bandDelay==1 && anular.bandDelay==1 && menique.bandDelay==1){//cuando todos los dedos terminan de corregir el delay se cumple esta condición
      bandDelay=1;
      preferences.putInt("bandDelay",bandDelay);//se guarda en preferences
      Serial.print("bandDelay: ");
      Serial.println(bandDelay);

      Serial.println("Delay antes de la corección:");
      Serial.print("delay open indice: ");
      Serial.println(indice.delayOpen);
      Serial.print("delay close indice: ");
      Serial.println(indice.delayClose);

      Serial.print("delay open medio: ");
      Serial.println(medio.delayOpen);
      Serial.print("delay close medio: ");
      Serial.println(medio.delayClose);

      Serial.print("delay open anular: ");
      Serial.println(anular.delayOpen);
      Serial.print("delay close anular: ");
      Serial.println(anular.delayClose);

      Serial.print("delay open meñique: ");
      Serial.println(menique.delayOpen);
      Serial.print("delay close meñique: ");
      Serial.println(menique.delayClose);

      correccionDelay(&indice,&medio,&anular,&menique);
    }
    
  }
 //se obtiene y se muestran los valores obtenidos en la función de calibración y corrección de los delay
  indice.potMin=preferences.getInt("potMinIndice",indice.potMin);
  indice.potMax=preferences.getInt("potMaxIndice",indice.potMax);
  indice.delayOpen=preferences.getInt("delayOpenInd",indice.delayOpen);
  indice.delayClose=preferences.getInt("delayCloseInd",indice.delayClose);
  Serial.print("delay open indice: ");
  Serial.println(indice.delayOpen);
  Serial.print("delay close indice: ");
  Serial.println(indice.delayClose);
  
  medio.potMin=preferences.getInt("potMinMedio",medio.potMin);
  medio.potMax=preferences.getInt("potMaxMedio",medio.potMax);
  medio.delayOpen=preferences.getInt("delayOpenMed",medio.delayOpen);
  medio.delayClose=preferences.getInt("delayCloseMed",medio.delayClose);
  Serial.print("delay open medio: ");
  Serial.println(medio.delayOpen);
  Serial.print("delay close medio: ");
  Serial.println(medio.delayClose);

  anular.potMin=preferences.getInt("potMinAnular",anular.potMin);
  anular.potMax=preferences.getInt("potMaxAnular",anular.potMax);
  anular.delayOpen=preferences.getInt("delayOpenAnu",anular.delayOpen);
  anular.delayClose=preferences.getInt("delayCloseAnu",anular.delayClose);
  Serial.print("delay open anular: ");
  Serial.println(anular.delayOpen);
  Serial.print("delay close anular: ");
  Serial.println(anular.delayClose);

  menique.potMin=preferences.getInt("potMinMenique",menique.potMin);
  menique.potMax=preferences.getInt("potMaxMenique",menique.potMax);
  menique.delayOpen=preferences.getInt("delayOpenMeniq",menique.delayOpen);
  menique.delayClose=preferences.getInt("delayCloseMeniq",menique.delayClose);
  Serial.print("delay open menique: ");
  Serial.println(menique.delayOpen);
  Serial.print("delay close menique: ");
  Serial.println(menique.delayClose);

}

void loop(){

  if(S_A==true && S_C==false && S_Q==false){
    abreDedo(&indice);
    abreDedo(&anular);
    abreDedo(&medio);
    abreDedo(&menique);
    if(indice.estadoDedo==1 && medio.estadoDedo==1 && anular.estadoDedo==1 && menique.estadoDedo==1){
      S_A=false;
      S_C=true;
      S_Q=true;
      Serial.println("termina de abrir");
      Serial.print("PWM indice: ");
      Serial.println(indice.pwm);

      Serial.print("PWM medio: ");
      Serial.println(medio.pwm);

      Serial.print("PWM anular: ");
      Serial.println(anular.pwm);

      Serial.print("PWM meñique: ");
      Serial.println(menique.pwm);
      /*Serial.print("PWM indice: ");
      Serial.println(indice.pwm);

      Serial.print("PWM medio: ");
      Serial.println(medio.pwm);

      Serial.print("PWM anular: ");
      Serial.println(anular.pwm);

      Serial.print("PWM meñique: ");
      Serial.println(menique.pwm);*/

    }
  }

  else if(S_C==true && S_A==false && S_Q==false){
    cierraDedo(&indice);
    cierraDedo(&anular);
    cierraDedo(&medio);
    cierraDedo(&menique);
    
    if(indice.estadoDedo==2 && medio.estadoDedo==2 && anular.estadoDedo==2 && menique.estadoDedo==2){
      S_A=true;
      S_C=false;
      S_Q=true;
      Serial.println("termina de cerrar");

      Serial.print("PWM indice: ");
      Serial.println(indice.pwm);

      Serial.print("PWM medio: ");
      Serial.println(medio.pwm);

      Serial.print("PWM anular: ");
      Serial.println(anular.pwm);

      Serial.print("PWM meñique: ");
      Serial.println(menique.pwm);
    }
  }

  else{
    dedoQuieto(&indice);
    dedoQuieto(&anular);
    dedoQuieto(&medio);
    dedoQuieto(&menique);
    delay(10000);// hay que cambiar delay por contador
    S_Q=false;
  }
// se actualizan los valores de PWM
  if(((S_A==true|| S_C==true)&& S_Q==false)&& (millis()-lastCalibration>callTime)){//1:abriendo, 2:cerrando, 3:quieto
    setPWM(&indice,&medio,&anular,&menique);
    //Serial.println("set PWM ");
    lastCalibration=millis();
  }
}